// lighting-related code:
// structures, uniforms for lights and material
// functions for light calculation
//

// CGUE: cook-torrance removed, only one simple directional light

// mat_diff_color and mat_spec_color are global vars to be set before calling any lighting funcs
// phong shading considers different material colors for diffuse/specular light

// material (Phong model)
struct Material {
	vec3 ambient, diffuse, specular;
	float shininess;
};

// lights
struct BasicLight {
	int  enabled;
	vec3 color_amb, color_diff, color_spec;
};

struct DirectionalLight {
	BasicLight light;
	vec3 direction;
};

struct Attenuation {
	float constant, linear, quadratic;
};

struct PointLight {
	BasicLight light;
	vec3 position;
	Attenuation atten;
};

struct SpotLight {
	PointLight pointlight;
	vec3 direction;
	float cos_cutoff1, cos_cutoff2; // cosine of inner/outer cutoff angle
};


// uniforms

#define MAX_DIRLIGHTS   1
//#define MAX_POINTLIGHTS 4
//#define MAX_SPOTLIGHTS  1

layout (location=8)  uniform vec3 eye;				// eye position (camera)
//layout (location=10) uniform int lightModel;		// lighting model (0=Phong, 1=Cook-Torrance)
layout (location=11) uniform Material material;		// material uniform Phong (loc. 11-15)
//layout (location=16) uniform MaterialCT materialct;	// material uniform Cook-Torrance (loc. 16-19)
layout (location=20) uniform DirectionalLight dirLight[MAX_DIRLIGHTS]; // 20..23  (4 per light)	
//layout (location=40) uniform PointLight pointLight[MAX_POINTLIGHTS];   // 40..63  (6 per light)		// note: uniform locations no longer used/valid!
//layout (location=70) uniform SpotLight spotLight[MAX_SPOTLIGHTS];      // 70..78  (9 per light)		// note: uniform locations no longer used/valid!

uniform float globalBrightnessMultiplier = 1.0;

// global vars for light functions
Material final_material;

vec3 accumulatedAmbientLight;
int  accumulatedAmbientLightCount;

// light functions

vec3 calcPhongLight(BasicLight light, vec3 vLight, vec3 vView, vec3 vNorm) {
	// calculate light intensity (without ambient) by the Phong model
	// vLight vector from point to light source
	// vView  vector from point to eye
	// vNorm  normal vector
	// all vectors normalized!

	vec3 vRefl = reflect(-vLight, vNorm); // reflected light beam vector
	
	// calc phong model factors
	float diff_factor = max(dot(vLight, vNorm), 0.0f); // clamp to 0
	float spec_factor = max(dot(vRefl,  vView), 0.0f); // clamp to 0

	vec3 res_d = final_material.diffuse  * diff_factor * light.color_diff;
	vec3 res_s;
	if (final_material.shininess == 0) {
		res_s = vec3(0); // no shininess; in some obj. files...
	} else {
		res_s = final_material.specular * pow(spec_factor, final_material.shininess) * light.color_spec;
	}
	
	accumulatedAmbientLight = accumulatedAmbientLight + light.color_amb;
	accumulatedAmbientLightCount++;

	return light.enabled * (res_d + res_s);
}


vec3 calcDirectionalLight(DirectionalLight dl, vec3 vView, vec3 vNorm) {
	// light vector (from frag to light) is just the reverse direction
	return calcPhongLight(dl.light, normalize(-dl.direction), vView, vNorm);
}

vec3 calcPointLight(PointLight pl, vec3 vView, vec3 vNorm, vec3 fragmentPos) {
	vec3 vLight = pl.position - fragmentPos;
	float d = length(vLight);
	float atten;
	atten = pl.atten.constant + d*pl.atten.linear + d*d*pl.atten.quadratic;
	return calcPhongLight(pl.light, normalize(vLight), vView, vNorm) / atten;
}

vec3 calcSpotLight(SpotLight sl, vec3 vView, vec3 vNorm, vec3 fragmentPos) {
	// note: cutoff1 = smaller angler = larger cos
	// cut off        if angle > cutoff2 ==> cosangle < cos_cutoff2
	// full intensity if angle < cutoff1 ==> cosangle > cos_cutoff1

	float intensity = 0.0;

	vec3 vLight = normalize(sl.pointlight.position - fragmentPos);
	float cosangle = max(dot(vLight, normalize(-sl.direction)), 0.0f);

	intensity = smoothstep(sl.cos_cutoff2, sl.cos_cutoff1, cosangle);
	return intensity * calcPointLight(sl.pointlight, vView, vNorm, fragmentPos);
}

// begin lighting calculations
void beginLighting() {
	accumulatedAmbientLightCount = 0;
	accumulatedAmbientLight = vec3(0.0);
}

// finish lighting calculations - returns final color
vec3 finishLighting(vec3 intensity) {
	vec3 f;
	if (accumulatedAmbientLightCount > 0) {
		f = accumulatedAmbientLight / accumulatedAmbientLightCount;
	} else {
		f = vec3(1.0);
	}
	return (intensity + f * final_material.ambient) * globalBrightnessMultiplier;
}

